#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <string.h>
#include <ctype.h>

#include "icfp.h"

void AllocError() {
	debug("Allocation Error!\n");
	exit(1);
}

int RobotWeight() {
	int i;
	int weight = 0;
	for (i = 0; i < MAX_PACKAGES; i++) {
		if (carrying[i].id >= 0)
			weight = weight + carrying[i].weight;
	}
	return weight;
}

int RobotX() {
	int i;
	for (i = 0; i < MAX_PLAYERS; i++) {
		if (players[i].id == robot.id) {
			return players[i].x;
		}
	}
	return -1;
}

int RobotY() {
	int i;
	for (i = 0; i < MAX_PLAYERS; i++) {
		if (players[i].id == robot.id) {
			return players[i].y;
		}
	}
	return -1;
}
	
void PrintNumberMap(int h_board[board.height][board.width]) {
	int i,j;
	if (!DEBUG)
		return;
	for (i = 0; i < board.height; i++) {
		for (j = 0; j < board.width; j++) {
			
			debug("\t%d",h_board[i][j]);
		}
		debug("\n");
	}
	debug("\n\n");
}

void PrintBoard() {
	int i, j;
	if (!DEBUG)
		return;
	debug("robot x / y:\t%d %d\n",RobotX(),RobotY());
	debug("score:\t\t%d\n",score);
	debug("cash:\t\t%d\n",robot.cash);
	debug("picked up:\t\t%d\n",picked_up_packages);
	debug("dropped:\t\t%d\n",dropped_packages);
	debug("weight:\t\t%d, max %d\n",RobotWeight(),robot.max_weight);
	debug("Board:\n");
	for (i = 0; i < board.height; i++) {
		debug("%d\t",i+1);
		for (j = 0; j < board.width; j++) {
			if (j+1 == RobotX() &&
			    i+1 == RobotY()) 
				debug(" R");
			else
				debug(" %c",board.loc[i][j]);
		}
		debug("\n");
	}
	debug("\n\t");
	for (i = 1; i < board.width+1; i++) {
		debug("%.2d",i);
	}
	debug("\n\n");
}

void CompressPackages() { /* Squeeeeze! :o) */
	int i;
	int empty_loc = 0;
	int loc = -1;
	for (i = 0; i < MAX_PACKAGES; i++) {
		if (carrying[i].id == -1) {
			empty_loc = 1;
			loc = i;
		}
		if (carrying[i].id >= 0 &&
		    carrying[i].id < MAX_PACKAGES &&
		    empty_loc == 1 &&
		    loc != -1) {
			carrying[loc].id = carrying[i].id;
			carrying[loc].x  = carrying[i].x;
			carrying[loc].y  = carrying[i].y;
			carrying[loc].weight = carrying[i].weight;
			carrying[i].id = -1;
			carrying[i].x = -1;
			carrying[i].y = -1;
			carrying[i].weight = -1;
			empty_loc = 0;
			loc = -1;
			i = 0;
		}
	}
}

void GetBoard() {
	int row_counter, col_counter;

/* Get Board Dimensions and allocate array*/
	ReadLine();
	sscanf(buffer,"%d %d\n",
	       &board.width,
	       &board.height);
	board.loc = (char **) malloc(board.width * board.height);
	if (board.loc == NULL) {
		AllocError();
	}
/* Read board.height lines and copy them into board.loc */
	for (row_counter = 0;
	     row_counter < board.height;
	     row_counter++) {
	  ReadLine();
		board.loc[row_counter] = (char *) malloc(board.width);
		if (board.loc[row_counter] == NULL) {
			AllocError();
		}
		for(col_counter = 0;
		    col_counter < board.width;
		    col_counter++) {
			if (buffer[col_counter] != '.' &&
			    buffer[col_counter] != '~' &&
			    buffer[col_counter] != '#' &&
			    buffer[col_counter] != '@') {
				debug("strange char: %c at %d %d\n",
				       buffer[col_counter],
				       row_counter,
				       col_counter);
			}
			board.loc[row_counter][col_counter] = buffer[col_counter];
		}
	}

}	

void GetConfig() {
	ReadLine();
	sscanf(buffer,"%d %d %d",
	       &robot.id,
	       &robot.max_weight,
	       &robot.cash);
}

void EndGame() {
	debug("* hostname was: %s at %d\n",server_hostname,server_port);
	debug("* robot statistics:\n*\tx = %d, y = %d\n*\trobot.id = %d\n*\trobot.cash = %d\n*\tScore: = %d\n\n",RobotX(),RobotY(),robot.id,robot.cash,score);
	/* Cleanup neaded here! */

	close(ssocket);
	exit(0);
}

void RobotDeath() {
	debug("AAAAARRRRAAARRRGGGGG!!!! Them Bastids got us sir!!!\n\n");
	EndGame();
}

void KillRobot(int i) {
	players[i].id = 0;
	players[i].x  = 0;
	players[i].y  = 0;
} 

void MoveNorth(int robot_id) {
	int i = 0;
	while (i < MAX_PLAYERS) {
		if (players[i].id == robot_id) {
			players[i].y++;
			break;
		}
		i++;
	}
}

void MoveSouth(int robot_id) {
	int i = 0;
	while (i < MAX_PLAYERS) {
		if (players[i].id == robot_id) {
			players[i].y--;
			break;
		}
		i++;
	}
}

void MoveEast(int robot_id) {
	int i = 0;
	while (i < MAX_PLAYERS) {
		if (players[i].id == robot_id) {
			players[i].x++;
			break;

		}
		i++;
	}
}

void MoveWest(int robot_id) {
	int i = 0;
	while (i < MAX_PLAYERS) {
		if (players[i].id == robot_id) {
			players[i].x--;
			break;

		}
		i++;
	}
}

void AddPackage(struct s_package package) {
	int i = 0;
	while(carrying[i].id >= 0) {
		i++;
	}
	carrying[i].id = package.id;
	carrying[i].x  = package.x;
	carrying[i].y  = package.y;
	carrying[i].weight = package.weight;
}

void PickupPackage(int robot_id, int package_id) {
	int i = 0;
	if (robot_id != robot.id) 
		return;
	for (i = 0; i < MAX_PACKAGES; i++) {
		if (packages[i].id == package_id) { 
			AddPackage(packages[i]);
			picked_up_packages++;
			return;
		}
	}
	return;
}

void DropPackage(int robot_id, int package_id) {
	int i = 0;
	if (robot_id != robot.id) { /* If not our robot, no direct consequences */
		return;
	}
	while(carrying[i].id != package_id) {
		i++;
		if (i > MAX_PACKAGES)
			return;
	}
	if (carrying[i].x == RobotX() &&
	    carrying[i].y == RobotY()) {
		/* Adding wieght to score */
		score = score + carrying[i].weight;
		dropped_packages++;
	}
	carrying[i].id = -1;
	carrying[i].x = -1;
	carrying[i].y = -1;
	carrying[i].weight = -1;
}

char *ProcesReply(char *pointer) {
	int i = 0;
	int valid = 0;
	int current_robot = -1;
	int package_id = -1;

	while(pointer[i] != '\n') {
		/* if (pointer[i] == '#' && current_robot != -1) {
				return &pointer[i];
		} */

		if (pointer[i] == '#') {
			sscanf(pointer,"#%d",&current_robot);
			i++;
			pointer = &pointer[i];
			i = 0;
			continue;
		}
		if (current_robot == -1) {
			debug("Error, robotcommand without id");
			i++;
			pointer = &pointer[i];
			i = 0;
			continue;
		}
		if (isdigit(pointer[i])) {
			i++;
			pointer = &pointer[i];
			i = 0;
			continue;
		}

		switch (pointer[i]) {
		case 'N':
			
			MoveNorth(current_robot);
			i++;
			pointer = &pointer[i];
			i = 0;
			break;
		case 'S':
			
			MoveSouth(current_robot);
			i++;
			pointer = &pointer[i];
			i = 0;
			break;
		case 'E':
			
			MoveEast(current_robot);
			i++;
			pointer = &pointer[i];
			i = 0;
			break;
		case 'W':
		
			MoveWest(current_robot);
			i++;
			pointer = &pointer[i];
			i = 0;
			break;
		case 'P':
			if (current_robot != robot.id) {
				i++;
				break;
			}
			sscanf(&pointer[i],"P %d",&package_id);
			PickupPackage(current_robot,package_id);
			i++;
			while (pointer[i] != '#' ||
			       pointer[i] != '\n') {
				while (pointer[i] != 'P' &&
				       pointer[i] != '#' &&
				       pointer[i] != 'D' &&
				       pointer[i] != 'N' &&
				       pointer[i] != 'W' &&
				       pointer[i] != 'E' &&
				       pointer[i] != 'S' &&
				       pointer[i] != '\n') {
					i++;
				}
				if (pointer[i] == '#' ||
				    pointer[i] == 'D' ||
				    pointer[i] == 'N' ||
				    pointer[i] == 'W' ||
				    pointer[i] == 'E' ||
				    pointer[i] == 'S' ||
				    pointer[i] == '\n') {
					pointer = &pointer[i];
					i = 0;
					break;
				}
				sscanf(&pointer[i],"P %d",&package_id);
				PickupPackage(current_robot,package_id);
				i++;
			}
			pointer = &pointer[i];
			i = 0;
			break;
		case 'D':
			if (current_robot != robot.id) {
				i++;
				break;
			}
			sscanf(&pointer[i],"D %d",&package_id);
			DropPackage(current_robot,package_id);
			i++;
			while (pointer[i] != '#' ||
			       pointer[i] != '\n') {
				while (pointer[i] != 'D' &&
				       pointer[i] != 'N' &&
				       pointer[i] != 'W' &&
				       pointer[i] != 'S' &&
				       pointer[i] != 'E' &&
				       pointer[i] != '#' &&
				       pointer[i] != '\n') {
					i++;
				}
				if (pointer[i] == 'N' ||
				    pointer[i] == 'W' ||
				    pointer[i] == 'S' ||
				    pointer[i] == 'E') { /*being pushed! */
					pointer = &pointer[i];
					i = 0;
					valid = 0;
					CompressPackages();
					break;
				}
				if (pointer[i] == '#' ||
				    pointer[i] == '\n') {
					pointer = &pointer[i];
					i = 0;
					CompressPackages();
					break;
				}
				sscanf(&pointer[i],"D %d",&package_id);
				DropPackage(current_robot,package_id);
				i++;
			}
			pointer = &pointer[i];
			i = 0;
			CompressPackages();
			break;
		case ' ':
			i++;
			pointer = &pointer[i];
			i = 0;
			break;
		case 'X':
			players[current_robot].id = current_robot;
			sscanf(&pointer[i],"X %d Y %d",
			       &players[current_robot].x,
			       &players[current_robot].y);
			
			i++;

			while(pointer[i] == ' ' ||
			      pointer[i] == 'X' ||
			      pointer[i] == 'Y' ||
			      isdigit(pointer[i])) {
				i++;
			}
			pointer = &pointer[i];
			i = 0;
			break;
		default:
			debug("Received strange command: %c\n",pointer[i]);
		}
	}
	return &pointer[i];
}

int GetPlayerPosition() {
	char *pointer;

	/* We dont know our new position or if we are still alive! */

	ReadLine();
	pointer = buffer;
	
	pointer = ProcesReply(pointer);
	
	if (RobotX() == -1 || RobotY() == -1) {
		/* Our poor robot wasn't found, c'est la vie... */
		RobotDeath();
	}

	return 1;
}

void ClearPlayers(){
	int i;
	for(i = 0; i < MAX_PLAYERS; i++) {
		players[i].id = -1;
		players[i].x = 0;
		players[i].y = 0;
	}
}

void GetPackageList() {
	int i = 0;
	int j = 0;
	int package_count = 0;
	char *pointer;

	ReadLine();
	pointer = buffer;
	while(pointer[i] != '\n') {
		if (isdigit(pointer[i])) {
			sscanf(&pointer[i],"%d %d %d %d",
			       &packages[package_count].id,
			       &packages[package_count].x,
			       &packages[package_count].y,
			       &packages[package_count].weight);
			
			j = 0;
			while(j < 3 || pointer[i] != ' ') {
				if(pointer[i] == ' ') {
					j++;
				}
				i++;
			}
			pointer = &pointer[i];
			package_count++;
			i = 0;
		}
		else {
			i++;
		}
	}
	if (package_count == 0) {
		debug("Alas, no packages here...");
	}
}  


char GetHighestAsChar(int above, int below, int left, int right) {
	int random;

	if (above > below &&
	    above > left &&
	    above > right) 
		return 'N';
	
	if (below > left &&
	    below > right &&
	    below > above)
		return 'S';

	if (left > right &&
	    left > below &&
	    left > above)
		return 'W';
	
	if (right > left &&
	    right > below &&
	    right > above)
		return 'E';

	if (above == below &&
	    above == right &&
	    above == left) {
		random = 1+(int) (4.0*rand()/(RAND_MAX+1.0));
		if (random == 1) 
			return 'N';
		if (random == 2)
			return 'S';
		if (random == 3)
			return 'W';
		return 'E';
	}

	if (above == below &&
	    above == right &&
	    above > left) {
		random = 1+(int) (3.0*rand()/(RAND_MAX+1.0));
		if (random == 1) 
			return 'N';
		if (random == 2)
			return 'S';
		return 'E';
	}
	if (above == below &&
	    above == left  &&
	    above > right) {
		random = 1+(int) (3.0*rand()/(RAND_MAX+1.0));
		if (random == 1) 
			return 'N';
		if (random == 2)
			return 'S';
		return 'W';
	}
	if (right == below &&
	    right == left  &&
	    right > above) {
		random = 1+(int) (3.0*rand()/(RAND_MAX+1.0));
		if (random == 1) 
			return 'E';
		if (random == 2)
			return 'S';
		return 'W';
	}
	if (above == below &&
	    above > left   &&
	    above > right) {
		random = 1+(int) (2.0*rand()/(RAND_MAX+1.0));
		if (random == 1) 
			return 'N';
		return 'S';
	}
	if (above == right &&
	    above > left  &&
	    above > below) {
		random = 1+(int) (2.0*rand()/(RAND_MAX+1.0));
		if (random == 1) 
			return 'N';
		return 'E';
	}
	if (above == left &&
	    above > right &&
	    above > below) {
		random = 1+(int) (2.0*rand()/(RAND_MAX+1.0));
		if (random == 1) 
			return 'N';
		return 'W';
	}
	if (below == right &&
	    below > left &&
	    below > above) {
		random = 1+(int) (2.0*rand()/(RAND_MAX+1.0));
		if (random == 1) 
			return 'E';
		return 'S';
	}
	if (below == left &&
	    below > right &&
	    below > above) {
		random = 1+(int) (2.0*rand()/(RAND_MAX+1.0));
		if (random == 1) 
			return 'W';
		return 'S';
	}
	if (left == right &&
	    left > above &&
	    left > below) {
		random = 1+(int) (2.0*rand()/(RAND_MAX+1.0));
		if (random == 1) 
			return 'W';
		return 'E';
	}
	/* shouldnt get here */
	return 'N';
}

int GetHighest(int above, int below, int left, int right) {
	char highest = GetHighestAsChar(above,below,left,right);
	switch (highest) {
	case 'N':
		return above;
	case 'S':
		return below;
	case 'W':
		return left;
	case 'E':
		return right;
	default: 
		return above;
	}
}

int AdjacentNumber(int h_board[board.height][board.width], int i, int j) {
	int above = -2, below = -2, left = -2, right = -2;
	
	if (i+1 < board.height) {
		if (h_board[i+1][j] > 0) {
			above = h_board[i+1][j];
		}
	}
	if (i-1 >= 0) {
		if (h_board[i-1][j] > 0) {
			below = h_board[i-1][j];
		}
	}
	if (j+1 < board.width) {
		if (h_board[i][j+1] > 0) {
			right = h_board[i][j+1];
		}
	}
	if (j-1 >= 0) {
		if (h_board[i][j-1] > 0) {
			left = h_board[i][j-1];
		}
	}
	if (above <= -1 &&
	    below <= -1 &&
	    left <= -1 &&
	    right <= -1)
		return -2; 

	return GetHighest(above,below,left,right) - 1;
}

int MovementRobotOpen(int h_board[board.height][board.width]) {
	int open = 0;
	int max_open = 0;
	int i = RobotY()-1;
	int j = RobotX()-1;

	if (i+1 < board.height) {
		max_open++;
		if (h_board[i+1][j] == -2) {
			open++;
		}
	}
	if (i-1 >= 0) {
		max_open++;
		if (h_board[i-1][j] == -2) {
			open++;
		}
	}
	if (j+1 < board.width) {
		max_open++;
		if (h_board[i][j+1] == -2) {
			open++;
		}
	}
	if (j-1 >= 0) {
		max_open++;
		if (h_board[i][j-1] == -2) {
			open++;
		}
	}
	if (open != 0)
		return 1;
	return 0;
}

char GetRobotHighest(int h_board[board.height][board.width]) {
	int above = -2, below = -2, left = -2, right = -2;
	int i = RobotY() - 1;
	int j = RobotX() - 1;
	
	if (i+1 < board.height) {
		if (h_board[i+1][j] > 0) {
			above = h_board[i+1][j];
			debug("above: %d\n",above);
		}
	}
	if (i-1 >= 0) {
		if (h_board[i-1][j] > 0) {
			below = h_board[i-1][j];
			
			debug("below: %d\n",below);
		}
	}
	if (j+1 < board.width) {
		if (h_board[i][j+1] > 0) {
			right = h_board[i][j+1];

			debug("right: %d\n",right);
		}
	}
	if (j-1 >= 0) {
		if (h_board[i][j-1] > 0) {
			left = h_board[i][j-1];

			debug("left: %d\n",left);
		}
	}
	return GetHighestAsChar(above,below,left,right);
}

char GetRandomDirection() { /* Random aint suicidal or stupid =) */
	int random;
	int i;
	while(1) {
		i++;
		random = 1+(int) (4.0*rand()/(RAND_MAX+1.0));
		if (random == 1 && RobotY() < board.height &&
		    board.loc[RobotY()][RobotX()-1] != '~' &&
		    board.loc[RobotY()][RobotX()-1] != '#') 
			return 'N';
		if (random == 2 && (RobotY()-2) >= 0 &&
		    board.loc[RobotY()-2][RobotX()-1] != '~' &&
		    board.loc[RobotY()-2][RobotX()-1] != '#')
			return 'S';
		if (random == 3 && RobotX() < board.width &&
		    board.loc[RobotY()-1][RobotX()] != '~' &&
		    board.loc[RobotY()-1][RobotX()] != '#')
			return 'E';
		if (random == 4 && (RobotX()-2) >= 0 &&
		    board.loc[RobotY()-1][RobotX()-2] != '~' &&
		    board.loc[RobotY()-1][RobotX()-2] != '#')
			return 'W';
		if (i > 1000)
			break;
	}

	/* Do stupid move before suicidal move! Rather Theo than dead! */
	i = 0;
	while(1) {
		i++;
	random = 1+(int) (4.0*rand()/(RAND_MAX+1.0));
	if (random == 1 && RobotY() < board.height &&
	    board.loc[RobotY()][RobotX()-1] != '~')
		return 'N';
	if (random == 2 && (RobotY()-2) >= 0 &&
	    board.loc[RobotY()-2][RobotX()-1] != '~')
		return 'S';
	if (random == 3 && RobotX() < board.width &&
	    board.loc[RobotY()-1][RobotX()] != '~')
		return 'E';
	if (random == 4 && (RobotX()-2) >= 0 &&
	    board.loc[RobotY()-1][RobotX()-2] != '~')
		return 'W';
	if (i > 1000)
		break;
	}

	if (RobotY() >= board.height)
		return 'N';
	if ((RobotY()-2) < 0)
		return 'S';
	if (RobotX() >= board.width)
		return 'E';
	if ((RobotY()-2) < 0)
		return 'W';
	/* Mommy... */
	return 'N';
}

char *FindBase() {
	char *action;
	int i, j, z, omega = 0;
	char adj;
	int alter = 0;
	int h_board[board.height][board.width];
	action = (char *)malloc(MAX_COMMAND_LENGTH);
	
	for (i = 0; i < board.height; i++) {
		for (j = 0; j < board.width; j++) {
			switch (board.loc[i][j]) {
			case '.':
				h_board[i][j] = -2;
				break;
			case '#':
				h_board[i][j] = -1;
				break;
			case '~':
				h_board[i][j] = -1;
				break;
			case '@':
				h_board[i][j] = MAX_HEUR_AT_BASE;
				for (z = 0; z < MAX_HOME_BASES; z++) {
					if (bases[z].x == -1)
						break;
					if (bases[z].x == j+1 &&
					    bases[z].y == i+1 &&
					    bases[z].standoff > 0) {
						h_board[i][j] = -1;
						bases[z].standoff--;
						break;
					}
						/* Bad? Worse? Probably!
		h_board[i][j] = MAX_HEUR_AT_BASE - bases[z].standoff--;
							*/
				}
				break;
			}
		}
	}
	while(MovementRobotOpen(h_board)) {
		omega++;
		if (alter == 0) {
			for (i = 0; i < board.height; i++) {
				for (j = 0; j < board.width; j++) {
					if (h_board[i][j] <= -2)
						h_board[i][j] = AdjacentNumber(h_board,i,j);
				}
			}
			alter = 1;
		}
		else {
			for (i = board.height - 1; i >= 0; i--) {
				for (j = board.width -1 ; j >= 0; j--) {
					if (h_board[i][j] <= -2)
						h_board[i][j] = AdjacentNumber(h_board,i,j);
				}
			}
			alter = 0;
		}
		if (omega > 10000) { /* High, but should be needed when base cant be found! */
			adj = GetRandomDirection();
			sprintf(action,"Move %c\n",adj);
			return action;
		}
		//PrintNumberMap(h_board);
	}
	adj = GetRobotHighest(h_board);

	sprintf(action,"Move %c\n",adj);
	return action;
}

int AtBase() {
	if (board.loc[RobotY()-1][RobotX()-1] == '@') {
		debug("Wohoo! At Base!");
		return 1;
	}
	return 0;
}

int AdjacentWater(int x, int y) {
	if((y < board.height && board.loc[y][x-1] == '~') ||
	   (y-2 > 0 && board.loc[y-2][x-1] == '~') ||
	   (x < board.width && board.loc[y-1][x] == '~') ||
	   (x-2 > 0 && board.loc[y-1][x-2] == '~'))
		return 1;
	else
		return 0;
}

int CloseToRobot() {
	int i;
	for (i = 0; i < MAX_PLAYERS; i++) {
		if (players[i].id < 0)
			continue;
		if (players[i].id == robot.id)
			continue;
		if ((abs(players[i].x - RobotX()) + abs(players[i].y - RobotY()))
		    < 3)
			return 1;
	}
	return 0;
}

int MakeBid() {
	if (AdjacentWater(RobotX(),RobotY()) && CloseToRobot()) {
		if ((robot.cash / 1000) > 5 &&
		    (robot.cash / 1000) < 30)
			return robot.cash / 1000;
		if ((robot.cash / 1000) > 30)
			return 30;
		return 5;
	}
		
	return 1;
}

int Chance(int chance) {
	if ((1+(int) (((float)chance)*rand()/(RAND_MAX+1.0))) == 1)
		return 1;
	else 
		return 0;
}

int RandomizeSign() {
	
	if ((1+(int) (2.0*rand()/(RAND_MAX+1.0))) == 1)
		return 1;
	else 
		return -1;
}

int GetVDistance(int id) {
	/* Very nice algo here, it just takes too long (15 secs, drat) =)
	for (i = 0; i < board.height; i++) {
		for (j = 0; j < board.width; j++) {
			if (packages[id].x-1 == j &&
			    packages[id].y-1 == i) {
				h_board[i][j] = 10000;
				continue;
			}
			switch (board.loc[i][j]) {
			case '.':
				h_board[i][j] = -2;
				break;
			case '#':
				h_board[i][j] = -1;
				break;
			case '~':
				h_board[i][j] = -1;
				break;
			case '@':
				h_board[i][j] = -2;
				break;
			}
		}
	}
	while(MovementRobotOpen(h_board)) {
		omega++;
		if (alter == 0) {
			for (i = 0; i < board.height; i++) {
				for (j = 0; j < board.width; j++) {
					if (h_board[i][j] <= -2)
						h_board[i][j] = AdjacentNumber(h_board,i,j);
				}
			}
			alter = 1;
		}
		else {
			for (i = board.height - 1; i >= 0; i--) {
				for (j = board.width -1 ; j >= 0; j--) {
					if (h_board[i][j] <= -2)
						h_board[i][j] = AdjacentNumber(h_board,i,j);
				}
			}
			alter = 0;
		}
		if (omega > 100) { 
			return -10000;
		}
	}
	*/
	return packages[id].weight * 10 - (abs(RobotX() - packages[id].x) * abs(RobotY() - packages[id].y));
}

int ReturnPackageWithID(int id) {
	int i;
	for(i = 0; i < MAX_PACKAGES; i++) {
		if(packages[i].id < 0 || packages[i].id > MAX_PACKAGES)
			break;
		if (packages[i].id == id) {
			return i;
		}
	}
	return -1;
}

char * PickUpPackages(int dest) {
	int i, j, z;
	char *action;
	char *temp;
	int package_list[MAX_PACKAGES];
	int number_packages = 0;
	int est_weight = 0;
	int v_distance = 0;
	int v_distance_2 = 0;
	action = (char *)malloc(MAX_COMMAND_LENGTH);

	for (i = 0; i < MAX_PACKAGES; i++) {
		if (packages[i].id < 0 || packages[i].id > MAX_PACKAGES)
			break;

		if (packages[i].weight + est_weight <= robot.max_weight) {
			/* Pick up package */
			package_list[number_packages] = packages[i].id;
			number_packages++;
			est_weight = est_weight + packages[i].weight;
		}

		else {
			/* Check to see if this package may be a better one... */
			v_distance = GetVDistance(i);
			for (j = 0; j < MAX_PACKAGES;j++) {
				if (packages[j].id < 0 || packages[j].id > MAX_PACKAGES)
					break;
					
				for (z = 0; z < number_packages; z++) {
					v_distance_2 = GetVDistance(ReturnPackageWithID(package_list[z]));
					if (v_distance > v_distance_2) {
						package_list[z] = packages[i].id;
						break;
					}
				}
			}
		}
	}
	if (number_packages == 0 && dest == 1) { /* No packages at base, use standoff*/
		for (i = 0; i < MAX_HOME_BASES; i++) {
			if (bases[i].x == RobotX() &&
			    bases[i].y == RobotY()) {
				bases[i].standoff = MAX_HOME_BASE_STANDOFF;
				break;
			}
			if (bases[i].x == -1) {
				bases[i].x = RobotX();
				bases[i].y = RobotY();
				bases[i].standoff = MAX_HOME_BASE_STANDOFF;
				break;
			}
		}
		return FindBase();
	}

	sprintf(action,"Pick");
	for (i = 0; i < number_packages; i++) {
		temp = (char *) malloc(10);
		sprintf(temp," %d",packages[i].id);
		action = strcat(action,temp);
	}
	action = strcat(action,"\n");
	return action;
}

int GetTotalHeuristicWeight(int x, int y) {
	int i;
	int h_weight = 0;
	
	for (i = 0; i < MAX_PACKAGES; i++) {
		if (carrying[i].id < 0 || carrying[i].id > MAX_PACKAGES)
			break;
		if (carrying[i].x == x &&
		    carrying[i].y == y) {
			h_weight+=carrying[i].weight;
		}
	}
	return h_weight;
}

char * FindDestination() {
	char *action;
	int i, j, omega = 0;
	char adj;
	int alter = 0;
	int h_weight = 0;
	int h_board[board.height][board.width];

	action = (char *)malloc(MAX_COMMAND_LENGTH);
	for (i = 0; i < board.height; i++) {
		for (j = 0; j < board.width; j++) {
			h_weight = GetTotalHeuristicWeight(1+j,1+i);

			switch (board.loc[i][j]) {
			case '.':
				if (h_weight > 0) {
					h_board[i][j] = 100*h_weight;
				}
				else {
					h_board[i][j] = -2;
				}
				break;
			case '#':
				h_board[i][j] = -1;
				break;
			case '~':
				h_board[i][j] = -1;
				break;
			case '@':
				if (h_weight > 0) {
					h_board[i][j] = 100*h_weight;
				}
				else {
					h_board[i][j] = -2;
				}
				break;
			}
		}
	}
	while(MovementRobotOpen(h_board)) {
		omega++;
		if (alter == 0) {
			for (i = 0; i < board.height; i++) {
				for (j = 0; j < board.width; j++) {
					if (h_board[i][j] <= -2)
						h_board[i][j] = AdjacentNumber(h_board,i,j);
				}
			}
			alter = 1;
		}
		else {
			for (i = board.height - 1; i >= 0; i--) {
				for (j = board.width -1 ; j >= 0; j--) {
					if (h_board[i][j] <= -2)
						h_board[i][j] = AdjacentNumber(h_board,i,j);
				}
			}
			alter = 0;
		}
		if (omega > 10000) { /* High, but should be needed when dest cant be found! */
			adj = GetRandomDirection();
			sprintf(action,"Move %c\n",adj);
			return action;
		}
	}
	adj = GetRobotHighest(h_board);

	sprintf(action,"Move %c\n",adj);
	return action;
}

char * DropAllPackages() {
	int i;
	char *action;
	char *temp;
	int package_list[MAX_PACKAGES];
	int number_packages = 0;
	action = (char *)malloc(MAX_COMMAND_LENGTH);

	for (i = 0; i < MAX_PACKAGES; i++) {
		if (carrying[i].id < 0 || carrying[i].id > MAX_PACKAGES)
			break;

		if (carrying[i].x == RobotX() &&
		    carrying[i].y == RobotY()) {
			/* Pick up package */
			package_list[number_packages] = carrying[i].id;
			number_packages++;
		}
	}
	sprintf(action,"Drop");
	for (i = 0; i < number_packages; i++) {
		temp = (char *) malloc(10);
		sprintf(temp," %d",carrying[i].id);
		action = strcat(action,temp);
	}
	action = strcat(action,"\n");
	return action;
}

int CheckPackages() {
	int i;
	for (i = 0; i < MAX_PACKAGES; i++) {
		if (carrying[i].id < 0 || carrying[i].id > MAX_PACKAGES)
			break;
		if (carrying[i].x == RobotX() &&
		    carrying[i].y == RobotY()) {
			return carrying[i].id;
		}
	}
	return -1;
}

int CheckPackagesGround() {
	int i;
	int number = 0;
	for (i = 0; i < MAX_PACKAGES; i++) {
		if (packages[i].id < 0 || packages[i].id > MAX_PACKAGES)
			break;
		number++;
	}
	return number;
}

char *MakeAction() {
	char *action;
	int package_id = -1;
	
	//action = (char *)malloc(MAX_COMMAND_LENGTH);
	
	package_id = CheckPackages();
	if (CloseToRobot() && Chance(15)) { /* Try to do something unpredictable =) */
		action = (char *)malloc(MAX_COMMAND_LENGTH);
		sprintf(action,"Move %c\n",GetRandomDirection());
		return action;
	}
	if (package_id == -1) {
		if (AtBase()) {
			if (RobotWeight() == 0) {
				if (CloseToRobot() && Chance(15)) {
					action = (char *)malloc(MAX_COMMAND_LENGTH);
					sprintf(action,"Move %c\n",GetRandomDirection());
				}
				else {
					action = PickUpPackages(1);
				}
			}
			else {
				action = FindDestination();
			}
		}
		else {
			if (RobotWeight() == 0) {
				if (CheckPackagesGround() > 0 && RandomizeSign()) {
					action = PickUpPackages(0);
				}
				else {
					action = FindBase();
				}
			}
			else {
				action = FindDestination();
			}
		}
	}
	else {
		action = DropAllPackages();
	}
	debug("\nCommand input: Move N == down, Move E == right\n");
	/* Read t/m \n */
	/* fgets(action,MAX_COMMAND_LENGTH,stdin); */
	return action;
}

char *MakeCommand() {
	int bid = MakeBid();
	char *action = MakeAction();
	robot.cash -= bid;
	bid = bid * RandomizeSign();
	sprintf(command,"%d %s",bid, action);
	return command;
}

void ClearPackageList() {
	int i;
	for (i = 0; i < MAX_PACKAGES; i++) {
		packages[i].id = -1;
	}
}

void PlayGame() {
	char *command;
	
	score = 0;
	picked_up_packages = 0;
	dropped_packages = 0;
	command = (char *) malloc(MAX_COMMAND_LENGTH+10);
	PrintBoard();
	while(1) { /* Exit when no packages or dead robot */
		ClearPackageList();
		GetPackageList();
		command = MakeCommand();
		/* Note that command should end with a \n! */
		WriteLine(command);

		GetPlayerPosition();

		PrintBoard();
	}
}

void ClearBaselist() {
	int i;
	for (i = 0; i < MAX_HOME_BASES; i++) {
		bases[i].x = -1;
		bases[i].y = -1;
		bases[i].standoff = -1;
	}
}

void ClearCarrylist() {
	int i;
	for (i = 0; i < MAX_PACKAGES; i++) {
		carrying[i].id = -1;
	}
}

void InitialiseGame() {
	debug("Initialising game...\n");
	WriteLine("Player\n");
	ClearCarrylist();
	ClearBaselist();
	ClearPlayers();
	debug("Getting Board\n");
	GetBoard();
	GetConfig();
	GetPlayerPosition();
}
